/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2007 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:RegistryServiceImpl.java 10372 2007-05-14 13:58:42Z sauthieg $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.services.bootstrap.osgi;

import java.io.File;
import java.util.ArrayList;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.List;
import java.util.Properties;
import org.objectweb.util.monolog.api.Logger;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Constants;
import org.osgi.framework.ServiceReference;
import org.osgi.service.cm.Configuration;
import org.osgi.service.cm.ConfigurationAdmin;
import org.osgi.service.cm.ConfigurationPlugin;
import org.ow2.jonas.configuration.ConfigurationManager;
import org.ow2.jonas.lib.bootstrap.JProp;
import org.ow2.jonas.lib.bootstrap.LoaderManager;
import org.ow2.jonas.lib.util.Log;
import org.ow2.jonas.properties.ServerProperties;

/**
 * JOnAS Configuration component.
 * @author Francois Fornaciari
 */
public class JOnASConfiguration implements ConfigurationManager, ConfigurationPlugin {

    /**
     * Reference to the {@link ConfigurationAdmin} service.
     */
    private ConfigurationAdmin configurationAdmin = null;

    /**
     * Property name for the last modification.
     */
    private static final String DATE = "date";

    /**
     * Property name for the service name.
     */
    private static final String JONAS_SERVICE = "jonas.service";

    private static final String JONAS_BOOTSTRAP = "jonas.bootstrap";

    /**
     * {@link ServerProperties} exposed as an OGGi service.
     */
    private JProp serverProperties = null;

    /**
     * OSGi {@link BundleContext}.
     */
    private BundleContext bc = null;

    /**
     * Mandatory JOnAS services.
     */
    private List<String> mandatoryServices = null;

    /**
     * JOnAS runnable state. While the optional service configuration updates have not been done, it isn't.
     */
    private boolean isRunnable = false;

    /**
     * Management loggers.
     */
    private static Logger logger = Log.getLogger(Log.JONAS_BOOTSTRAP_PREFIX);

    /**
     * Constructor used to retrieve the OSGi BundleContext.
     * @param bc OSGi BundleContext
     */
    public JOnASConfiguration(final BundleContext bc) {
        this.bc = bc;

        // Read the JOnAS properties
        serverProperties = new JProp();
        serverProperties.setOSGi(true);

        // Get the OSGiLoaderManager instance
        LoaderManager lm = LoaderManager.getInstance(serverProperties);
        lm.init(serverProperties);

        // Initialize mandatory services list
        mandatoryServices = new ArrayList<String>();
        mandatoryServices.add("registry");
        mandatoryServices.add("jmx");
    }

    /**
     * Method called when the Configuration Admin Service is bound to the component. <br />
     * This method creates service configurations if they don't already exist or updates it if the JOnAS configuration
     * file has changed since the last modification.
     * @throws Exception Thrown if the configuration creation failed.
     */
    public void configure() throws Exception {
        // Retrieve the bootstrap system property
        String bootstrapMode = System.getProperty(JONAS_BOOTSTRAP);

        if (doUpdateServiceConfigurations()) {
            // Create only configurations for mandatory services if we are in bootstrap mode
            if (Boolean.parseBoolean(bootstrapMode)) {
                manageServiceConfigurations(mandatoryServices, false);
            } else {
                manageServiceConfigurations(getAllServices(), false);
            }
        }

        // The ServerProperties service must be registered after the configuration updates
        // to avoid that services are started before they are configured
        ServiceReference sr = bc.getServiceReference(ServerProperties.class.getName());
        if (sr == null) {
            bc.registerService(ServerProperties.class.getName(), serverProperties, null);
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void updateServiceConfigurations() throws Exception {
        if (doUpdateServiceConfigurations() || !isRunnable) {
            manageServiceConfigurations(getOptionalServices(), false);
            // The server is in a runnable state
            isRunnable = true;
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void deleteServiceConfigurations() throws Exception {
        manageServiceConfigurations(getOptionalServices(), true);
        // The server is no more in a runnable state
        isRunnable = false;
    }

    /**
     * {@inheritDoc}
     */
    public void haltServer() throws Exception {
        // Get the server configuration and delete it
        Configuration serverConfiguration = configurationAdmin.getConfiguration("org.ow2.jonas.server");
        serverConfiguration.delete();

        // Stop the system bundle
        bc.getBundle(0).stop();
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void updateServiceConfigurations(final List<String> services) throws Exception {
        manageServiceConfigurations(services, false);
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void deleteServiceConfigurations(final List<String> services) throws Exception {
        manageServiceConfigurations(services, true);
    }


    /**
     * Delete all service configurations.
     */
    private void deleteAllServiceConfigurations() throws Exception {
        manageServiceConfigurations(getAllServices(), true);

        // Get the server configuration and delete it
        Configuration serverConfiguration = configurationAdmin.getConfiguration("org.ow2.jonas.server");
        serverConfiguration.delete();
    }

    /**
     * Get 'true' if service configurations have to be updated, ie the server configuration don't already exist or the
     * file jonas.properties has not been modified since the last configuration update.
     * @return a <code>boolean</code> value to determine if service configurations have to be updated
     * @throws Exception Thrown either during the configuration creation or the configuration update
     */
    private boolean doUpdateServiceConfigurations() throws Exception {
        // Get the server configuration
        Configuration serverConfiguration = configurationAdmin.getConfiguration("org.ow2.jonas.server");
        // Get the server properties
        Dictionary serverProps = serverConfiguration.getProperties();

        // Init the server properties
        if (serverProps == null) {
            serverProps = new Properties();

            // Update the server configuration with the last modification time
            serverProps.put(DATE, getLastModification());
            serverConfiguration.update(serverProps);

            return true;
        } else {
            // Get time of the last modification of the jonas.properties file stored during the configuration update
            long currentModification = (Long) serverProps.get(DATE);
            // Time of the last modification of the jonas.properties file
            long lastModification = getLastModification();

            // If times are different, JOnAS properties file contains more recent information than the stored
            // OSGi configuration
            if (currentModification != lastModification) {
                // Read the last JOnAS properties
                serverProperties = new JProp();

                // Update the server configuration with the last modification time
                serverProps.put(DATE, lastModification);
                serverConfiguration.update(serverProps);

                return true;
            }
        }
        return false;
    }

    /**
     * Create, update or delete service configurations for the given services.
     * @param services List of services to configure
     * @param delete 'true' when configurations have to be delete
     * @throws Exception If an exception occurs during configuration updates or deletions
     */
    private void manageServiceConfigurations(final List<String> services, final boolean delete) throws Exception {
        for (String service : services) {
            boolean mandatory = mandatoryServices.contains(service);
            String serviceClass = JONAS_SERVICE + "." + service + ".class";
            String factoryPID = serverProperties.getValue(serviceClass);

            // Get the service properties
            Dictionary newProperties = null;

            if (!delete) {
                newProperties = getServiceProperties(service);

                // Add static properties
                newProperties.put(JONAS_SERVICE, service);
                newProperties.put("jonas.service.mandatory", mandatory);
            }

            // Get the list of configurations for a factory
            Configuration[] storedConfigurations = configurationAdmin.listConfigurations("("
                    + ConfigurationAdmin.SERVICE_FACTORYPID + "=" + factoryPID + ")");

            Configuration configuration = null;

            // There is no stored configuration for this factory
            if (!delete && storedConfigurations.length == 0) {
                // Create a configuration for this factory PID
                configuration = configurationAdmin.createFactoryConfiguration(factoryPID, null);

                // Update the configuration
                configuration.update(newProperties);
            } else {
                // A configuration already exists
                configuration = configurationAdmin.getConfiguration(storedConfigurations[0].getPid());

                if (delete) {
                    configuration.delete();
                } else {
                    // Look if service properties have been changed
                    Dictionary storedProperties = configuration.getProperties();
                    if (isConfigurationModified(storedProperties, newProperties)) {
                        // Update the configuration
                        configuration.update(newProperties);
                    }
                }
            }
        }
    }

    /**
     * Get all services.
     * @return All services
     */
    private List<String> getAllServices() {
        List<String> allServices = getOptionalServices();

        // Add mandatory services
        for (int i = 0; i < mandatoryServices.size(); i++) {
            allServices.add(i, mandatoryServices.get(i));
        }

        return allServices;
    }

    /**
     * Get optional services.
     * @return Optional services
     */
    private List<String> getOptionalServices() {
        // Get list of all declared services
        String[] declaredServices = serverProperties.getValueAsArray("jonas.services");
        List<String> optionalServices = new ArrayList<String>();

        // Add optional services ignoring mandatory services
        for (String declaredService : declaredServices) {
            if (!mandatoryServices.contains(declaredService)) {
                optionalServices.add(declaredService);
            }
        }

        return optionalServices;
    }

    /**
     * Get 'true' if the two dictionaries don't contain the same properties.
     * @param storedProperties Stored properties
     * @param newProperties New properties for JOnAS properties
     * @return A <code>boolean</code> value representing the fact that the two dictionaries contain or not the same
     *         properties
     */
    private boolean isConfigurationModified(final Dictionary storedProperties, final Dictionary newProperties) {
        // Remove automatic properties added by the Configuration Admin service
        storedProperties.remove(Constants.SERVICE_PID);
        storedProperties.remove(ConfigurationAdmin.SERVICE_FACTORYPID);
        storedProperties.remove(ConfigurationAdmin.SERVICE_BUNDLELOCATION);

        if (storedProperties.size() != newProperties.size()) {
            return true;
        } else {
            for (Enumeration enProps = storedProperties.keys(); enProps.hasMoreElements();) {
                String key = (String) enProps.nextElement();

                if (newProperties.get(key) == null) {
                    return true;
                } else if (!storedProperties.get(key).equals(newProperties.get(key))) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Get all properties for a given service name.
     * @param service Service name
     * @return A <code>Dictionary</code> value representing all properties for a given service name
     */
    private Dictionary getServiceProperties(final String service) {
        Dictionary props = new Properties();
        for (Object key : serverProperties.getConfigFileEnv().keySet()) {
            String name = (String) key;
            String servicePrefix = JONAS_SERVICE + "." + service + ".";
            if (name.startsWith(servicePrefix)) {
                String value = serverProperties.getValue(name);
                name = name.substring(servicePrefix.length());
                props.put(name, value);
            }
        }
        return props;
    }

    /**
     * Returns the time that the jonas.properties file was last modified.
     * @return A <code>long</code> value representing the time the file was last modified
     */
    private long getLastModification() {
        File conf = new File(JProp.getJonasBase(), "conf");
        File jonasPropertyFile = new File(conf, "jonas.properties");
        return jonasPropertyFile.lastModified();
    }

    /**
     * {@inheritDoc}
     */
    public void modifyConfiguration(final ServiceReference serviceReference, final Dictionary storedProperties) {
        try {
            if (doUpdateServiceConfigurations()) {
                String serviceName = (String) storedProperties.get(JONAS_SERVICE);

                // Get the last service properties
                Dictionary newProperties = getServiceProperties(serviceName);

                // Set new properties
                for (Enumeration enProps = newProperties.keys(); enProps.hasMoreElements();) {
                    String key = (String) enProps.nextElement();
                    storedProperties.put(key, newProperties.get(key));
                }

                manageServiceConfigurations(getAllServices(), false);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
