/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: JgroupsDiscoveryServiceImpl.java 12401 2007-12-14 14:17:47Z eyindanga $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.discovery.jgroups;

import java.util.ArrayList;
import java.util.HashMap;

import javax.management.InstanceNotFoundException;
import javax.management.JMException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.remote.JMXServiceURL;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.jonas.discovery.base.BaseDiscovery;
import org.ow2.jonas.discovery.jgroups.comm.exception.StopDiscException;
import org.ow2.jonas.discovery.jgroups.manager.DiscoveryManager;
import org.ow2.jonas.discovery.jgroups.utils.JGroupsDiscoveryUtils;
import org.ow2.jonas.jmx.JmxService;
import org.ow2.jonas.lib.util.JonasObjectName;
import org.ow2.jonas.lib.util.Log;
import org.ow2.jonas.properties.ServerProperties;
import org.ow2.jonas.service.ServiceException;
/**
 * This is a JGroups based JOnAS discovery implementation.
 * @author   eyindanga
 */
public class JgroupsDiscoveryServiceImpl extends BaseDiscovery implements JgroupsDiscoveryServiceImplMBean {

    /**
     * Logger for this service.
     */
    private static Logger logger = Log.getLogger(Log.JONAS_DISCOVERY_PREFIX);

    /**
     *  This property is injected.
     */
    private String groupName;
    /**
     *  This property is injected.
     */
    private String conf;
    /**
     * The discovery Manager.
     */
    private DiscoveryManager dm = null;


    /* (non-Javadoc)
     * @see org.ow2.jonas.lib.service.AbsServiceImpl#doStart()
     */
    @SuppressWarnings("unchecked")
    @Override
    public void doStart() throws ServiceException {
        String discoveryType = JGroupsDiscoveryUtils.DISCOVERY_IS_SLAVE;
        isDiscoveryMaster = getServerProperties().isMaster();
        if (isDiscoveryMaster) {
            discoveryType = JGroupsDiscoveryUtils.DISCOVERY_IS_MASTER;
        }
        logger = Log.getLogger(Log.JONAS_DISCOVERY_PREFIX);
        String domainName = jmxService.getDomainName();
        JMXServiceURL[] connectorServerURLs = jmxService
        .getConnectorServerURLs();
        urlsList = new ArrayList();
        for (int i = 0; i < connectorServerURLs.length; i++) {
            // The connectorServerURLs may contain null
            // if the list of protocols in Carol contain
            // other protocols than the standard ones (JRMP, IIOP, CMI)
            if (connectorServerURLs[i] != null) {
                urlsList.add(connectorServerURLs[i].toString());
            }
        }
        String[] urls = new String[urlsList.size()];
        for (int i = 0; i < urls.length; i++) {
            urls[i] = (String) urlsList.get(i);
        }
        try {
            //init configuration
            JGroupsDiscoveryUtils.init(jmxService.getJonasServerName(), jmxService.getDomainName(),
                    discoveryType, conf, groupName, jmxService.getJmxServer(), urls);

        } catch (Throwable e) {
            throw new ServiceException(e.getMessage());
        }
        startDomainMonitor(discoveryType);
        // Create and register the service MBean
        try {
            jmxService.registerMBean(this, JonasObjectName.discoveryService(domainName));
        } catch (Exception e) {
            logger.log(BasicLevel.DEBUG, " Unable to register discovery MBean: ",e);
        }
        logger.log(BasicLevel.INFO, "JGroups based Discovery service started from "+conf+" Stack file \n");

    }


    /**
     * Initializes and starts domain monitor.
     * @param discoveryType MASTER, SLAVE or CLUSTERD
     * @throws ServiceException if the service cannot be started.
     */
    private void  startDomainMonitor(final String discoveryType) throws ServiceException {
        ttl = JGroupsDiscoveryUtils.getInstance().getTtl();
        dm = new DiscoveryManager(discoveryType);
        // Start discovery manager
        try {
            dm.start();
        } catch (StopDiscException e) {
            // Server with the same name already exists in domain.
            logger.log(BasicLevel.DEBUG,
                    "Discovery manager failed to start due to a pre-existing"
                    + " server in the domain with the same name.", e);
            // Discovery service had a problem.
            throw new ServiceException(
                    "Problem when starting the Discovery Service: ", e);
        }
    }

    /** Starts JGroups discovery with a given environment and a mbean server.
     * @param env Contains discovery properties
     * @param mbeanSrv Contains discovery properties
     * @param discoveryType MBean server of discovery host.
     * @throws Exception Any
     */
    @SuppressWarnings("unchecked")
    public void start(final HashMap<String, Object> env, final MBeanServer mbeanSrv, final String discoveryType) throws Exception{
        if (discoveryType.equals(JGroupsDiscoveryUtils.DISCOVERY_IS_CLUSTERD)) {
            logger.log(BasicLevel.DEBUG, "Discovery started by cluster daemon");
            isDiscoveryMaster = false;
            groupName =(String) env.get("group.name");
            conf = (String) env.get("jgroups.conf");
            String hostName = (String) env.get("host.name");
            String[] urls = null;
            if (this.urlsList == null) {
                this.urlsList = new ArrayList<String>();
                try {
                    urls =(String[]) env.get("connector.urls");
                    for (int i = 0; i < urls.length; i++) {
                        urlsList.add(urls[i]);
                    }
                } catch (Exception e) {
                    logger.log(BasicLevel.DEBUG, " Exception occurred while getting connector urls for host named "+hostName+" in JGrpups discovery \n" );
                    throw new Exception(e.getMessage() + " Exception occurred while getting connector urls for host named "+hostName+" in JGrpups discovery \n");
                }

            }
            String domainName = (String) env.get("domain.name");
            try {
                JGroupsDiscoveryUtils.init(hostName, domainName, JGroupsDiscoveryUtils.DISCOVERY_IS_CLUSTERD, conf, groupName, mbeanSrv, urls);
            } catch (Throwable e) {
                throw new ServiceException(e.getMessage());
            }
            startDomainMonitor(discoveryType);

        }else {
            logger.log(BasicLevel.DEBUG, " JOnAS instance should not use this method");
        }
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.lib.service.AbsServiceImpl#doStop()
     */
    @Override
    public void doStop() throws ServiceException {
        //TODO Stop Clusterd discovery.
        dm.stop();
        if (JGroupsDiscoveryUtils.isRegistered(JonasObjectName.discoveryService(JGroupsDiscoveryUtils.getInstance().getDomainName()))) {
             logger.log(BasicLevel.DEBUG, "Discovery MBean is registered in the host MBean server\n");
             try {
                 JGroupsDiscoveryUtils.unRegisterMBean(JonasObjectName.discoveryService(JGroupsDiscoveryUtils.getInstance().getDomainName()));
             } catch (InstanceNotFoundException e) {
                 throw new ServiceException(e.getMessage());
             } catch (MBeanRegistrationException e) {
                 throw new ServiceException(e.getMessage());
             }
        }
        logger.log(BasicLevel.INFO, " Discovery service stopped");
    }
    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.DiscoveryServiceImplMBean#getDiscoveryProtocolVersion()
     */
    public String getDiscoveryProtocolVersion() {
        return JGroupsDiscoveryUtils.DISCOVERY_PROTOCOL_VERSION;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.DiscoveryServiceImplMBean#getDiscoveryTtl()
     */
    public String getDiscoveryTtl() {
        return (new Integer(ttl)).toString();
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.DiscoveryServiceImplMBean#startDiscoveryMaster()
     */
    public void startDiscoveryMaster() throws JMException {

    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#getJonasName()
     */
    public String getJonasName() {
        return JGroupsDiscoveryUtils.getInstance().getJonasName();
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#getMbeanServer()
     */
    public MBeanServer getMbeanServer() {
        return jmxService.getJmxServer();
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#getMyOn()
     */
    public ObjectName getMyOn() {
        return JonasObjectName.discoveryService(JGroupsDiscoveryUtils.getInstance().getDomainName());
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#getServerId()
     */
    public String getServerId() {
        return JGroupsDiscoveryUtils.getInstance().getServerId();
    }
    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#getUrls()
     */
    public String[] getUrls() {
        return (String[])this.urlsList.toArray();
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#setDomainName(java.lang.String)
     */
    public void setDomainName(final String domainName) {
        JGroupsDiscoveryUtils.getInstance().setDomainName(domainName);
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#setJonasName(java.lang.String)
     */
    public void setJonasName(final String jonasName) {
        JGroupsDiscoveryUtils.getInstance().setJonasName(jonasName);
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#setMbeanServer(javax.management.MBeanServer)
     */
    public void setMbeanServer(final MBeanServer mbeanServer) {
        JGroupsDiscoveryUtils.getInstance().setMbeanServer(mbeanServer);
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#setMyOn(javax.management.ObjectName)
     */
    public void setMyOn(final ObjectName myOn) {

    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#setServerId(java.lang.String)
     */
    public void setServerId(final String serverId) {
        JGroupsDiscoveryUtils.getInstance().setServerId(serverId);
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#setTtl(int)
     */
    public void setTtl(final int ttl) {
        this.ttl = ttl;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#setUrls(java.lang.String[])
     */
    public void setUrls(final String[] urls) {
        this.urlsList = toArrayList(urls);

    }
    /**
     * Build and Arraylist from String[] containing jmx urls.
     * @param urls
     * @return
     * @throws NullPointerException
     */
    @SuppressWarnings("unchecked")
    private ArrayList toArrayList(final String[] urls) throws NullPointerException {
        ArrayList<String> ret = new ArrayList();
        try {
            for (int i = 0; i < urls.length; i++) {
                ret.add(urls[i]);
            }
        } catch (NullPointerException e) {
            throw new NullPointerException("NullPointerException occurred in JgroupsDiscoveryServiceImpl. Urls list must not be null in toArrayList method \n"+e);
        }
        return ret;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.lib.service.AbsServiceImpl#start()
     */
    public void start() throws ServiceException {
        this.doStart();
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.lib.service.AbsServiceImpl#stop()
     */
    public void stop() {
        try {
            doStop();
        } catch (Exception e) {
            // TODO: handle exception
             logger.log(BasicLevel.ERROR, "Unable to stop discovery service \n"+e);
        }
         logger.log(BasicLevel.INFO, "Discovery servive stopped \n");

    }
    /**
     * @return The groupName
     */
    public String getGroupName() {
        return groupName;
    }

    /**
     * @param groupName the groupName to set. Used by the injector
     */
    public void setGroupName(final String groupName) {
        this.groupName = groupName;
    }

    /**
     * @return The JGroups configuration file name
     */
    public String getJgroupsConf() {
        return conf;
    }

    /**
     * @param conf The JGroups configuration file name to set. Used by the injector
     */
    public void setJgroupsConf(final String conf) {
        this.conf = conf;
    }


    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#setMaster(boolean)
     */
    @Override
    public void setMaster(final boolean isMaster) {
        this.isDiscoveryMaster = isMaster;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#getGreetingTimeOut()
     */
    @Override
    public int getGreetingTimeout() {
        return greetingAckTimeOut;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#getJmxService()
     */
    @Override
    public JmxService getJmxService() {
        return jmxService;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#getListeningIp()
     */
    @Override
    public String getListeningIp() {
        return JGroupsDiscoveryUtils.getInstance().getLocalAddress().getIpAddress().toString();
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#getListeningPort()
     */
    @Override
    public int getListeningPort() {
        return JGroupsDiscoveryUtils.getInstance().getDiscPort();
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#getTtl()
     */
    @Override
    public int getTtl() {
        return ttl;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#getUrlsList()
     */
    @SuppressWarnings("unchecked")
    @Override
    public ArrayList getUrlsList() {
        return urlsList;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#isDiscoveryMaster()
     */
    @Override
    public boolean isDiscoveryMaster() {
        return new Boolean(isDiscoveryMaster);
    }
    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#setJmxService(org.ow2.jonas.jmx.JmxService)
     */
    @Override
    public void setJmxService(final JmxService jmxService) {
        this.jmxService = jmxService;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#setListeningIp(java.lang.String)
     */
    @Override
    public void setListeningIp(final String listeningIp) {
    //TODO Implement this.
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#setListeningPort(int)
     */
    @Override
    public void setListeningPort(final int listeningPort) {
        //TODO Implement this.
    }
    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.BaseDiscovery#setUrlsList(java.util.ArrayList)
     */
    @SuppressWarnings("unchecked")
    @Override
    public void setUrlsList(final ArrayList urlsList) {
        logger.log(BasicLevel.DEBUG, "Setting urls \n");
        this.urlsList = urlsList;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#getGreetingListeningPort()
     */
    public int getGreetingListeningPort() {
        logger.log(BasicLevel.DEBUG, "Getting listening port \n");
        return JGroupsDiscoveryUtils.getInstance().getDiscPort();
    }


    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#setTimeToLive(int)
     */
    public void setTimeToLive(final int ttl) {
        logger.log(BasicLevel.DEBUG, "Setting ttl: not yet implemented \n");
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.jgroups.JgroupsDiscoveryServiceImplMBean#setGreetingTimeOut(int)
     */
    public void setGreetingTimeout(final int greetingAckTimeOut) {
        logger.log(BasicLevel.DEBUG, "Setting greeting timeout \n");
        this.greetingAckTimeOut = greetingAckTimeOut;
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.DiscoveryServiceImplMBean#getIsDiscoveryMaster()
     */
    public Boolean getIsDiscoveryMaster() {
        logger.log(BasicLevel.DEBUG, "Checking if discovery is master \n");
        return new Boolean(isDiscoveryMaster);
    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.DiscoveryServiceImplMBean#getMulticastAddress()
     */
    public String getMulticastAddress() {
            return JGroupsDiscoveryUtils.getMulticastAddress();

    }

    /* (non-Javadoc)
     * @see org.ow2.jonas.discovery.base.DiscoveryServiceImplMBean#getMulticastPort()
     */
    public String getMulticastPort() {
        return new Integer(JGroupsDiscoveryUtils.getMulticastPort()).toString();
    }

    public String isMulticast() {
        return new Boolean(JGroupsDiscoveryUtils.getInstance().isMulticast()).toString();
    }
}
